@@ -93,7 +93,7 @@ static ValuePairSpec rfc3164[] =
   { "PROGRAM"  },
   { "PID"      },
   { "MESSAGE"  },
-  { "DATE", "R_DATE" },
+  { "DATE" },
   { 0 },
 };
 
@@ -198,11 +198,12 @@ vp_pairs_foreach(gpointer data, gpointer user_data)
   const LogTemplateOptions *template_options = ((gpointer *)user_data)[6];
   SBTHGString *sb = sb_th_gstring_acquire();
   VPPairConf *vpc = (VPPairConf *)data;
+  gint time_zone_mode = GPOINTER_TO_INT (((gpointer *)user_data)[7]);
 
   sb->type_hint = vpc->template->type_hint;
   log_template_append_format((LogTemplate *)vpc->template, msg,
                              template_options,
-                             LTZ_LOCAL, seq_num, NULL, sb_th_gstring_string(sb));
+                             time_zone_mode, seq_num, NULL, sb_th_gstring_string(sb));
 
   if (sb_th_gstring_string(sb)->len == 0)
     {
@@ -229,8 +230,8 @@ vp_msg_nvpairs_foreach(NVHandle handle, gchar *name,
     return FALSE;
 
   inc = (name[0] == '.' && (vp->scopes & VPS_DOT_NV_PAIRS)) ||
-    (name[0] != '.' && (vp->scopes & VPS_NV_PAIRS)) ||
-    (log_msg_is_handle_sdata(handle) && (vp->scopes & VPS_SDATA));
+        (name[0] != '.' && (vp->scopes & VPS_NV_PAIRS)) ||
+        (log_msg_is_handle_sdata(handle) && (vp->scopes & (VPS_SDATA + VPS_RFC5424)));
 
   for (j = 0; j < vp->patterns_size; j++)
     {
@@ -250,26 +251,31 @@ vp_msg_nvpairs_foreach(NVHandle handle, gchar *name,
   return FALSE;
 }
 
-/* runs over a set of ValuePairSpec structs and merges them into the value-pair set */
+static gboolean
+vp_find_in_set(ValuePairs *vp, gchar *name, gboolean exclude)
+{
+  gint j;
+  gboolean included = exclude;
+
+  for (j = 0; j < vp->patterns_size; j++)
+    {
+      if (g_pattern_match_string(vp->patterns[j]->pattern, name))
+        included = vp->patterns[j]->include;
+    }
+
+  return included;
+}
+
 static void
-vp_merge_set(ValuePairs *vp, LogMessage *msg, gint32 seq_num, ValuePairSpec *set, GTree *dest, const LogTemplateOptions *template_options)
+vp_merge_other_set(ValuePairs *vp, LogMessage *msg, gint32 seq_num, gint time_zone_mode, ValuePairSpec *set, GTree *dest, const LogTemplateOptions *template_options, gboolean exclude)
 {
   gint i;
   SBTHGString *sb;
 
   for (i = 0; set[i].name; i++)
     {
-      gint j;
-      gboolean exclude = FALSE;
-
-      for (j = 0; j < vp->patterns_size; j++)
-        {
-          if (g_pattern_match_string(vp->patterns[j]->pattern, set[i].name))
-            exclude = !vp->patterns[j]->include;
-        }
-
-      if (exclude)
-	continue;
+      if (!vp_find_in_set(vp, set[i].name, exclude))
+        continue;
 
       sb = sb_th_gstring_acquire();
 
@@ -277,7 +283,7 @@ vp_merge_set(ValuePairs *vp, LogMessage *msg, gint32 seq_num, ValuePairSpec *set
         {
         case VPT_MACRO:
           log_macro_expand(sb_th_gstring_string(sb), set[i].id, FALSE,
-                           template_options, LTZ_LOCAL, seq_num, NULL, msg);
+                           template_options, time_zone_mode, seq_num, NULL, msg);
           break;
         case VPT_NVPAIR:
           {
@@ -302,6 +308,20 @@ vp_merge_set(ValuePairs *vp, LogMessage *msg, gint32 seq_num, ValuePairSpec *set
     }
 }
 
+/* runs over the all macros and merges the selected ones by the pattern into the value-pair set */
+static void
+vp_merge_macros(ValuePairs *vp, LogMessage *msg, gint32 seq_num, gint time_zone_mode, GTree *dest, const LogTemplateOptions *template_options)
+{
+  vp_merge_other_set(vp, msg, seq_num, time_zone_mode, all_macros, dest, template_options, FALSE);
+}
+
+/* runs over a set of ValuePairSpec structs and merges them into the value-pair set */
+static void
+vp_merge_set(ValuePairs *vp, LogMessage *msg, gint32 seq_num, gint time_zone_mode, ValuePairSpec *set, GTree *dest, const LogTemplateOptions *template_options)
+{
+  vp_merge_other_set(vp, msg, seq_num, time_zone_mode, set, dest, template_options, TRUE);
+}
+
 static gboolean
 vp_foreach_helper (const gchar *name, const SBTHGString *hinted_value,
                    gpointer data)
@@ -324,13 +344,13 @@ vp_data_free (SBTHGString *s)
 gboolean
 value_pairs_foreach_sorted (ValuePairs *vp, VPForeachFunc func,
                             GCompareDataFunc compare_func,
-                            LogMessage *msg, gint32 seq_num,
+                            LogMessage *msg, gint32 seq_num, gint time_zone_mode,
                             const LogTemplateOptions *template_options,
                             gpointer user_data)
 {
   gpointer args[] = { vp, func, msg, GINT_TO_POINTER (seq_num), user_data, NULL,
                       /* remove constness, we are not using that pointer non-const anyway */
-                      (LogTemplateOptions *) template_options
+                      (LogTemplateOptions *) template_options, GINT_TO_POINTER(time_zone_mode)
                     };
   gboolean result = TRUE;
   gpointer helper_args[] = { func, user_data, &result };
@@ -344,22 +364,25 @@ value_pairs_foreach_sorted (ValuePairs *vp, VPForeachFunc func,
   /*
    * Build up the base set
    */
-  if (vp->scopes & (VPS_NV_PAIRS + VPS_DOT_NV_PAIRS + VPS_SDATA) ||
+  if (vp->scopes & (VPS_NV_PAIRS + VPS_DOT_NV_PAIRS + VPS_SDATA + VPS_RFC5424) ||
       vp->patterns_size > 0)
     nv_table_foreach(msg->payload, logmsg_registry,
                      (NVTableForeachFunc) vp_msg_nvpairs_foreach, args);
 
+  if (vp->patterns_size > 0)
+    vp_merge_macros(vp, msg, seq_num, time_zone_mode, scope_set, template_options);
+
   if (vp->scopes & (VPS_RFC3164 + VPS_RFC5424 + VPS_SELECTED_MACROS))
-    vp_merge_set(vp, msg, seq_num, rfc3164, scope_set, template_options);
+    vp_merge_set(vp, msg, seq_num, time_zone_mode, rfc3164, scope_set, template_options);
 
   if (vp->scopes & VPS_RFC5424)
-    vp_merge_set(vp, msg, seq_num, rfc5424, scope_set, template_options);
+    vp_merge_set(vp, msg, seq_num, time_zone_mode, rfc5424, scope_set, template_options);
 
   if (vp->scopes & VPS_SELECTED_MACROS)
-    vp_merge_set(vp, msg, seq_num, selected_macros, scope_set, template_options);
+    vp_merge_set(vp, msg, seq_num, time_zone_mode, selected_macros, scope_set, template_options);
 
   if (vp->scopes & VPS_ALL_MACROS)
-    vp_merge_set(vp, msg, seq_num, all_macros, scope_set, template_options);
+    vp_merge_set(vp, msg, seq_num, time_zone_mode, all_macros, scope_set, template_options);
 
   /* Merge the explicit key-value pairs too */
   g_ptr_array_foreach(vp->vpairs, (GFunc)vp_pairs_foreach, args);
@@ -374,23 +397,31 @@ value_pairs_foreach_sorted (ValuePairs *vp, VPForeachFunc func,
 
 gboolean
 value_pairs_foreach(ValuePairs *vp, VPForeachFunc func,
-                    LogMessage *msg, gint32 seq_num,
+                    LogMessage *msg, gint32 seq_num, gint time_zone_mode,
                     const LogTemplateOptions *template_options,
                     gpointer user_data)
 {
   return value_pairs_foreach_sorted(vp, func, (GCompareDataFunc) strcmp,
-                                    msg, seq_num, template_options, user_data);
+                                    msg, seq_num, time_zone_mode, template_options, user_data);
 }
 
 typedef struct
 {
-  GTrashStack stackp;
   gchar *key;
   gchar *prefix;
   gint prefix_len;
 
   gpointer data;
-} vp_walk_stack_t;
+} vp_walk_stack_data_t;
+
+#define VP_STACK_INITIAL_SIZE 16
+
+typedef struct
+{
+  vp_walk_stack_data_t **buffer;
+  guint buffer_size;
+  guint count;
+} vp_stack_t;
 
 typedef struct
 {
@@ -399,60 +430,116 @@ typedef struct
   VPWalkValueCallbackFunc process_value;
 
   gpointer user_data;
-  vp_walk_stack_t *stack;
+  vp_stack_t *stack;
 } vp_walk_state_t;
 
-static vp_walk_stack_t *
-vp_walker_stack_unwind_until (vp_walk_stack_t **stack, vp_walk_state_t *state,
-                              const gchar *name)
+static vp_stack_t *
+vp_stack_create()
 {
-  vp_walk_stack_t *t;
-
-  if (!stack)
-    return NULL;
-  if (!*stack)
-    return NULL;
-  if (strncmp(name, (*stack)->prefix, (*stack)->prefix_len) == 0)
-    return *stack;
-
-  while ((t = g_trash_stack_pop((GTrashStack **)stack)) != NULL)
-    {
-      if (strncmp(name, t->prefix, t->prefix_len) != 0)
-        {
-          vp_walk_stack_t *p = g_trash_stack_peek((GTrashStack **)stack);
-
-          if (p)
-            state->obj_end(t->key, t->prefix, &t->data,
-                           p->prefix, &p->data,
-                           state->user_data);
-          else
-            state->obj_end(t->key, t->prefix, &t->data,
-                           NULL, NULL,
-                           state->user_data);
-          g_free(t->key);
-          g_free(t->prefix);
-          g_free(t);
-        }
-      else
-        {
-          /* This one matched, put it back, PUT IT BACK! */
-          g_trash_stack_push((GTrashStack **)stack, t);
-          break;
-        }
-    }
-
-  return *stack;
+  vp_stack_t *stack = g_new(vp_stack_t, 1);
+  stack->buffer = g_new(vp_walk_stack_data_t *, VP_STACK_INITIAL_SIZE);
+  stack->buffer_size = VP_STACK_INITIAL_SIZE;
+  stack->count = 0;
+  return stack;
 }
 
 static void
-vp_walker_stack_unwind_all(vp_walk_stack_t **stack,
+vp_stack_destroy(vp_stack_t *stack)
+{
+  g_free(stack->buffer);
+  g_free(stack);
+}
+
+static void
+vp_stack_realloc(vp_stack_t *stack, guint new_size)
+{
+  g_assert(new_size > stack->buffer_size);
+  stack->buffer = g_renew(vp_walk_stack_data_t *, stack->buffer, new_size);
+  stack->buffer_size = new_size;
+}
+
+static void
+vp_stack_push(vp_stack_t *stack, vp_walk_stack_data_t *data)
+{
+  if (stack->count >= stack->buffer_size)
+    vp_stack_realloc(stack, stack->buffer_size * 2);
+
+  stack->buffer[stack->count++] = data;
+}
+
+static vp_walk_stack_data_t *
+vp_stack_peek(vp_stack_t *stack)
+{
+  if (stack->count == 0)
+    return NULL;
+
+  return stack->buffer[stack->count-1];
+}
+
+static vp_walk_stack_data_t *
+vp_stack_pop(vp_stack_t *stack)
+{
+  vp_walk_stack_data_t *data = NULL;
+
+  if (stack->count == 0)
+    return NULL;
+
+  data = stack->buffer[stack->count-1];
+  stack->count--;
+  return data;
+}
+
+static guint
+vp_stack_height(vp_stack_t *stack)
+{
+  return stack->count;
+}
+
+static void
+vp_walker_stack_unwind_until (vp_stack_t *stack, vp_walk_state_t *state,
+                              const gchar *name)
+{
+  vp_walk_stack_data_t *t;
+
+  if (!stack)
+    return;
+
+  while ((t = vp_stack_pop(stack)) != NULL)
+    {
+      vp_walk_stack_data_t *p;
+
+      if (strncmp(name, t->prefix, t->prefix_len) == 0)
+        {
+          /* This one matched, put it back, PUT IT BACK! */
+          vp_stack_push(stack, t);
+          break;
+        }
+
+      p = vp_stack_peek(stack);
+
+      if (p)
+        state->obj_end(t->key, t->prefix, &t->data,
+                       p->prefix, &p->data,
+                       state->user_data);
+      else
+        state->obj_end(t->key, t->prefix, &t->data,
+                       NULL, NULL,
+                       state->user_data);
+      g_free(t->key);
+      g_free(t->prefix);
+      g_free(t);
+    }
+}
+
+static void
+vp_walker_stack_unwind_all(vp_stack_t *stack,
                            vp_walk_state_t *state)
 {
-  vp_walk_stack_t *t;
+  vp_walk_stack_data_t *t;
 
-  while ((t = g_trash_stack_pop((GTrashStack **)stack)) != NULL)
+  while ((t = vp_stack_pop(stack)) != NULL)
     {
-      vp_walk_stack_t *p = g_trash_stack_peek((GTrashStack **)stack);
+      vp_walk_stack_data_t *p = vp_stack_peek(stack);
 
       if (p)
         state->obj_end(t->key, t->prefix, &t->data,
@@ -469,41 +556,101 @@ vp_walker_stack_unwind_all(vp_walk_stack_t **stack,
     }
 }
 
-static vp_walk_stack_t *
-vp_walker_stack_push (vp_walk_stack_t **stack,
+static vp_walk_stack_data_t *
+vp_walker_stack_push (vp_stack_t *stack,
                       gchar *key, gchar *prefix)
 {
-  vp_walk_stack_t *nt = g_new(vp_walk_stack_t, 1);
+  vp_walk_stack_data_t *nt = g_new(vp_walk_stack_data_t, 1);
 
   nt->key = key;
   nt->prefix = prefix;
   nt->prefix_len = strlen(nt->prefix);
   nt->data = NULL;
 
-  g_trash_stack_push((GTrashStack **)stack, nt);
+  vp_stack_push(stack, nt);
   return nt;
 }
 
+static void
+vp_walker_name_value_split_add_name_token(GPtrArray *array, const gchar *name,
+                                          int *current_name_start_idx,
+                                          int *index)
+{
+  gchar *token;
+
+  token = g_strndup(name + *current_name_start_idx, *index - *current_name_start_idx);
+  *current_name_start_idx = ++(*index);
+  g_ptr_array_add(array, (gpointer) token);
+}
+
+static GPtrArray *
+vp_walker_name_value_split(const gchar *name)
+{
+  int i;
+  int current_name_start_idx = 0;
+  GPtrArray *array = g_ptr_array_new();
+  size_t name_len = strlen(name);
+
+  for (i = 0; i < name_len; i++)
+    {
+      if (name[i] == '@')
+        {
+          i++;
+          while (g_ascii_isdigit(name[i]) || (name[i] == '.' && g_ascii_isdigit(name[i + 1])))
+            i++;
+        }
+      if (name[i] == '.')
+        vp_walker_name_value_split_add_name_token(array, name, &current_name_start_idx, &i);
+    }
+  if (current_name_start_idx <= i - 1)
+    vp_walker_name_value_split_add_name_token(array, name, &current_name_start_idx, &i);
+
+  if (array->len == 0)
+  {
+    g_ptr_array_free(array, TRUE);
+    return NULL;
+  }
+
+  return array;
+}
+
 static gchar *
-vp_walker_name_split(vp_walk_stack_t **stack, vp_walk_state_t *state,
+vp_walker_name_combine_prefix(GPtrArray *tokens, gint until)
+{
+  SBGString *s = sb_gstring_acquire();
+  gchar *str;
+  gint i;
+
+  for (i = 0; i < until; i++)
+    {
+      g_string_append(sb_gstring_string(s), g_ptr_array_index(tokens, i));
+      g_string_append_c(sb_gstring_string(s), '.');
+    }
+  g_string_append(sb_gstring_string(s), g_ptr_array_index(tokens, until));
+
+  str = g_strdup(sb_gstring_string(s)->str);
+
+  sb_gstring_release(s);
+
+  return str;
+}
+
+static gchar *
+vp_walker_name_split(vp_stack_t *stack, vp_walk_state_t *state,
                      const gchar *name)
 {
-  gchar **tokens, *key = NULL;
-  guint token_cnt, i, start;
+  GPtrArray *tokens;
+  gchar *key = NULL;
+  guint i, start;
 
-  tokens = g_strsplit(name, ".", 0);
-  token_cnt = g_strv_length(tokens);
+  tokens = vp_walker_name_value_split(name);
 
-  start = g_trash_stack_height((GTrashStack **)stack);
-  for (i = start; i < token_cnt - 1; i++)
+  start = vp_stack_height(stack);
+  for (i = start; i < tokens->len - 1; i++)
     {
-      gchar *next = tokens[i + 1];
-      vp_walk_stack_t *nt, *p = g_trash_stack_peek((GTrashStack **)stack);
-
-      tokens[i + 1] = NULL;
-      nt = vp_walker_stack_push(stack, g_strdup(tokens[i]),
-                                g_strjoinv(".", tokens));
-      tokens[i + 1] = next;
+      vp_walk_stack_data_t *p = vp_stack_peek(stack);
+      vp_walk_stack_data_t *nt = vp_walker_stack_push(stack, g_strdup(g_ptr_array_index(tokens, i)),
+                                 vp_walker_name_combine_prefix(tokens, i));
 
       if (p)
         state->obj_start(nt->key, nt->prefix, &nt->data,
@@ -516,8 +663,10 @@ vp_walker_name_split(vp_walk_stack_t **stack, vp_walk_state_t *state,
 
   /* The last token is the key (well, second to last, last being
      NULL), so treat that normally. */
-  key = g_strdup(tokens[token_cnt - 1]);
-  g_strfreev(tokens);
+  key = g_strdup(g_ptr_array_index(tokens, tokens->len - 1));
+
+  g_ptr_array_foreach(tokens, (GFunc)g_free, NULL);
+  g_ptr_array_free(tokens, TRUE);
 
   return key;
 }
@@ -527,18 +676,20 @@ value_pairs_walker(const gchar *name, TypeHint type, const gchar *value,
                    gpointer user_data)
 {
   vp_walk_state_t *state = (vp_walk_state_t *)user_data;
-  vp_walk_stack_t *st = g_trash_stack_peek((GTrashStack **)&state->stack);
+  vp_walk_stack_data_t *data;
   gchar *key;
   gboolean result;
 
-  st = vp_walker_stack_unwind_until (&st, state, name);
+  vp_walker_stack_unwind_until (state->stack, state, name);
+  key = vp_walker_name_split (state->stack, state, name);
+  data = vp_stack_peek (state->stack);
 
-  key = vp_walker_name_split (&st, state, name);
+  key = vp_walker_name_split (state->stack, state, name);
 
-  if (st)
-    result = state->process_value(key, st->prefix,
+  if (data != NULL)
+    result = state->process_value(key, data->prefix,
                                   type, value,
-                                  &st->data,
+                                  &data->data,
                                   state->user_data);
   else
     result = state->process_value(key, NULL,
@@ -548,9 +699,6 @@ value_pairs_walker(const gchar *name, TypeHint type, const gchar *value,
 
   g_free(key);
 
-  /* And the new stack becomes whatever we have in st now. */
-  state->stack = st;
-
   return result;
 }
 
@@ -565,7 +713,7 @@ value_pairs_walk(ValuePairs *vp,
                  VPWalkCallbackFunc obj_start_func,
                  VPWalkValueCallbackFunc process_value_func,
                  VPWalkCallbackFunc obj_end_func,
-                 LogMessage *msg, gint32 seq_num,
+                 LogMessage *msg, gint32 seq_num, gint time_zone_mode,
                  const LogTemplateOptions *template_options,
                  gpointer user_data)
 {
@@ -576,14 +724,15 @@ value_pairs_walk(ValuePairs *vp,
   state.obj_start = obj_start_func;
   state.obj_end = obj_end_func;
   state.process_value = process_value_func;
-  state.stack = NULL;
+  state.stack = vp_stack_create();
 
   state.obj_start(NULL, NULL, NULL, NULL, NULL, user_data);
   result = value_pairs_foreach_sorted(vp, value_pairs_walker,
                                       (GCompareDataFunc)vp_walk_cmp, msg,
-                                      seq_num, template_options, &state);
-  vp_walker_stack_unwind_all(&state.stack, &state);
+                                      seq_num, time_zone_mode, template_options, &state);
+  vp_walker_stack_unwind_all(state.stack, &state);
   state.obj_end(NULL, NULL, NULL, NULL, NULL, user_data);
+  vp_stack_destroy(state.stack);
 
   return result;
 }
